/****************************************************************************
 *
 * qn
 *
 * This file is part of libcppsim
 *
 * Copyright (C) 2004 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

/*
 * Queuing network simulator for acyclic networks with finite capacity and
 * BAS blocking and multiple classes.
 */
#include <queue>
#include <string>
#include <vector>
#include <iostream>
#include <cstdio>
#include <functional>
#include <algorithm>
#include "cppsim.hh"

extern "C" {
#include <getopt.h>
}

#include "qn_source.hh"
#include "qn_server.hh"
#include "qn_job.hh"
#include "qn_topology.hh"
#include "qn_param.hh"

using namespace std;
using namespace cppsim;

qnet* network;
qn_param_struct qn_param;

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Declarations
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

//
// class Simulation main
//
class simulation_main : public process {
public:
    simulation_main( const string& name ) : 
	process( name ) { };
protected:
    void inner_body( void );
};

vector< handle<server> > servers;


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Implementations
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//
// Class simulation_main
//
void simulation_main::inner_body( void )
{
    // Build servers
    servers.resize( network->nservers() );

    for ( int i=0; i<network->nservers(); i++ ) {
        servers[i] = new server( oform( "server_%d", i ), i );
        servers[i]->activateAfter( current() );
    }

    // Build sources    
    for ( int c=0; c<network->nclasses(); c++ ) {
        for ( int j=0; j<network->nservers(); j++ ) {
            if ( network->arrivals(c,j) > 0 ) {
                handle_p s( new source( oform( "source_%d_%d", j, c), c, 
                                        network->arrivals(c,j), j ) );
                s->activateAfter( current() );
            }
        }
    }

    // Start!
    hold( qn_param.simLen );

    // Finish simulation
    end_simulation( );
}

//
// Returns true if the simulation results converged
//
bool converged( void )
{
    cout << endl << endl;
    for ( int i=0; i<qn_param.queue_length.size(); i++ ) {
        try {
            confInt c = qn_param.queue_length[i]->value();
            cout << "qlen_" << i << "\t" 
                 << c.lBound() << " " << c.uBound() << "\t"
                 << (c.uBound()-c.lBound())/c.val() << endl;
            if ( (c.uBound() - c.lBound())/c.val() > qn_param.confw)
                return false;
        } catch ( runtime_error& e ) {
            return false;
        }
    }
    return true;
}

//
// Initialize all the statistics
//
void initialize_variables( void )
{
    const int bw = 200;
    qn_param.queue_length.resize( network->nservers() );
    qn_param.utilization.resize( network->nservers() );
    qn_param.realutilization.resize( network->nservers() );
    qn_param.queue_p.resize( network->nservers() );
    qn_param.throughput.resize( network->nservers() );

    for ( int i=0; i<network->nservers(); i++ ) {
        qn_param.queue_length[i] = 
            new timeweighted( oform("q_length_%d", i ), 
                              qn_param.confp, bw );

        qn_param.utilization[i] = 
            new timeweighted( oform("util_%d", i ), 
                              qn_param.confp, bw );

        qn_param.realutilization[i] = 
            new timeweighted( oform("realutil_%d", i ), 
                              qn_param.confp, bw );

        qn_param.queue_p[i] = 
            new timewhisto( oform( "P_%d", i ),
                            ( network->capacity(i) > 0 ?
                              network->capacity(i)+1 : 4 ),
                            qn_param.confp, bw );

        qn_param.throughput[i] = 
            new mean( oform( "X_%d", i ), qn_param.confp );        
    }
}

void reset_statistics( void )
{
    for_each( qn_param.queue_length.begin(), qn_param.queue_length.end(),
              mem_fun( &timeweighted::reset ) );
    for_each( qn_param.utilization.begin(), qn_param.utilization.end(),
              mem_fun( &timeweighted::reset ) );
    for_each( qn_param.realutilization.begin(), qn_param.realutilization.end(),
              mem_fun( &timeweighted::reset ) );
    for_each( qn_param.queue_p.begin(), qn_param.queue_p.end(),
              mem_fun( &timewhisto::reset ) );

    for ( int i=0; i<network->nservers(); i++ ) {
        qn_param.throughput[i]->update( servers[i]->completed() / qn_param.simLen );
    }
}

//
// Do the simulation experiment
//
void run_experiment( void )
{
    unsigned const int step = 5;
    unsigned const int limit = qn_param.maxRep;
    int i=0, l=0;
    bool conv = false;

    cout << "Replication ";

    while ( l < limit ) {
        l += step;
        for ( ; i<l; i++ ) {        
            
            if ( 0 == i % 5 )
                cout << i << flush;
            else
                cout << "." << flush;
            
            simulation *s = new simpleSimulation();
            handle<simulation_main> 
                sim_main1( new simulation_main( "Simulation main1" ) );
            sim_main1->activate( );
            s->run();
            delete s;
            
            // Reset statistics, if necessary
            if ( i < l - 1 ) 
                reset_statistics( );
        }
        if ( conv = converged() ) 
            break;
        else
            reset_statistics( );
    }
    if ( conv ) 
        cout << "done (converged)" << endl;
    else
        cout << "done (replication limit reached)" << endl;
}

//
// Print simulation results
//
void print_results( void )
{
    int i;

    try {
        for ( i=0; i<qn_param.queue_length.size(); i++ ) {
            confInt c( qn_param.queue_length[i]->value( ) );
            cout << "L_" << i << " " << c.lBound( ) << " " << c.uBound( ) 
                 << endl;
            vector< string > vh( qn_param.queue_p[i]->headers() );
            vector< string > vv( qn_param.queue_p[i]->entries() );
            for ( int j=0; j<vv.size(); j++ ) {
                cout << '\t' << vh[j] << ' ' << vv[j] << '\n';
            }

        }

        for ( i=0; i<qn_param.utilization.size(); i++ ) {
            confInt c( qn_param.utilization[i]->value( ) );
            cout << "U_" << i << " " << c.lBound( ) << " " << c.uBound( ) 
                 << endl;
        }

        for ( i=0; i<qn_param.realutilization.size(); i++ ) {
            confInt c( qn_param.realutilization[i]->value( ) );
            cout << "RU_" << i << " " << c.lBound( ) << " " << c.uBound( ) 
                 << endl;
        }
        for ( i=0; i<qn_param.throughput.size(); i++ ) {
            confInt c( qn_param.throughput[i]->value( ) );
            cout << "X_" << i << " " << c.lBound( ) << " " << c.uBound( ) 
                 << endl;
        }
    } catch ( runtime_error& e ) {
        cerr << "Runtime Error Caught: " << e.what() << endl;
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Main function
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
int main( int argc, char* argv[] )
{
    INIT_SEQUENCING;
    char opt;
    string fname;

    qn_param.maxRep     = 100;
    qn_param.confp      = 0.9;
    qn_param.confw      = 0.05;
    qn_param.simLen     = 2e4;
    qn_param.doDebug    = false;
    qn_param.trLen      = 0.2;

    /*
     * Read the command line parameters
     */
    while ( -1 != ( opt = getopt( argc, argv, "c:hr:l:t:Dw:" ) ) ) {
	switch (opt) {
	case 'c':
	    qn_param.confp = atof( optarg );
	    break;
	case 'h':
	    cout << "Usage: " << argv[0] 
		 << " [-c <confp>] [-h] [-r <numrep>] [-l <siml>] [-t <trlen>] [-D] [-S] <parameter file name>" << endl
		 << endl
		 << " -c <confp>  Sets the confidence probability to <confp> (def: " << qn_param.confp << ")" << endl
                 << " -w <confw>  Sets the confidence interval relative width (def: " << qn_param.confw << ")" << endl
		 << " -r <numrep> Sets the number of replications to <numrep> (def: " << qn_param.maxRep << ")" << endl
		 << " -l <siml>   Sets the length of the simulation, in time units (def: " << qn_param.simLen << ")" << endl
		 << " -t <trlen>  Length of the transient period (def: " << qn_param.trLen << ")" << endl
		 << " -D          Turns on lots of debugging messages (def off)" << endl
		 << " -h          Displays this help" << endl
		 << endl;
	    exit( 0 );
	    break;
	case 'r':
	    qn_param.maxRep = atoi( optarg );
	    assert( qn_param.maxRep >= 2 );
	    break;
	case 'l':
	    qn_param.simLen = atof( optarg );
	    break;
        case 'w':
            qn_param.confw = atof( optarg );
            break;
	case 't':
	    qn_param.trLen = atof( optarg );
	    break;
	case 'D':
	    qn_param.doDebug = true;
	    process::getLogger()->enable( logger::LOG_INFO | logger::LOG_DEBUG );
	    break;
	default:
	    cerr << "Unrecognized option: " << argv[ optind ] << endl;
	    exit( -1 );
	}
	
    }

    trremovalBuilderFrac myb( qn_param.trLen );
    trremovalDefaultBuilder = &myb;

    fname = string( argv[ optind ] );

    // 
    // Read the network
    //
    network = new qnet( fname );
    network->dump();

    initialize_variables( );

    run_experiment( );

    print_results( );

    return 0;
}
