/****************************************************************************
 *
 * qn_server.cc
 *
 * This file is part of libcppsim
 *
 * Copyright (C) 2004 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#include <string>

#include "qn_server.hh"
#include "qn_job.hh"
#include "qn.hh"
#include "qn_topology.hh"
#include "qn_param.hh"

using namespace std;
using namespace cppsim;

//
// Class server
//
server::server( const string& name, int serverno ) :
    process( name ),
    _serverno( serverno ),
    _capacity( network->capacity(serverno) ),
    _jobs( ),
    _blocked( ),
    _status( server::IDLE ),
    _servTime( network->nclasses() ),
    _nextServer( network->nclasses() ),
    _utilization( qn_param.utilization[serverno] ),
    _realutilization( qn_param.realutilization[serverno] ),
    _queue_length( qn_param.queue_length[serverno] ),
    _queue_p( qn_param.queue_p[serverno] ),
    _completed( 0 )
{ 
    // cout << "Server " << serverno << endl;
    for ( int c=0; c<network->nclasses(); c++ ) {
        // Init service times
        _servTime[c] = new rngExp( oform("rng_%d_rng", serverno),
                                   1.0/network->service( c,serverno ) );
        // cout << "\tclass " << c << " service rate " << network->service( c, serverno ) << endl;
        // Init routing probabilities
        vector< double > pr( network->nservers() );
        // cout << "\tclass " << c << " routing ";
        for ( int j=0; j<network->nservers(); j++ ) {
            pr[j] = network->topology(c, serverno, j );
            // cout << pr[j] << " ";
        }
        // cout << endl;
        _nextServer[c] = new rngDiscEmpirical( name+"_rng", pr );
    }
}

server::~server( )
{
    for ( int c=0; c<network->nclasses(); c++ ) {
        delete _servTime[c];
        delete _nextServer[c];
    }
    while ( !_jobs.empty() ) {
        job* j( _jobs.front( ) );
        _jobs.pop( );
        delete j;
    }
}

void server::add_new_job( job* j )
{
    assert( ( _capacity < 0 ) || ( _jobs.size( ) < _capacity ) );
    _jobs.push( j );
    _queue_length->update( make_pair( _jobs.size(), time() ) );
    //
    // Uncomment the next line if you want to compute the "true" queue
    // length. The definition of queue length does not include the job
    // currently being served.
    //
    // _queue_length->update( make_pair( _jobs.size()-1, time() ) );
    _queue_p->update( make_pair( _jobs.size(), time() ) );
}

void server::remove_first_job( void )
{
    assert( _jobs.size( ) > 0 );
    _jobs.pop( );
    _queue_length->update( make_pair( _jobs.size(), time() ) );
    //
    // Uncomment the next line if you want to compute the "true" queue
    // length. The definition of queue length does not include the job
    // currently being served.
    //
    // _queue_length->update( make_pair( ( _jobs.size() > 0 ? _jobs.size()-1 : 0 ), time() ) );
    _queue_p->update( make_pair( _jobs.size(), time() ) );
}

void server::add_waiting_proc( const handle_p& p ) 
{
    _blocked.push( p );
}

// void server::dump_status( void )
// {
//     for ( int i=0; i<network->nservers(); i++ ) {
//         handle<server> s( servers[i] );
//         message( this, "Server[%d] [%d/%d]\n", i, s->qlen(), s->maxlen() );
//     }
// }

void server::setStatus( status_t s )
{
    switch( s ) {
    case server::IDLE:
        _utilization->update( make_pair( 0.0, time() ) );
        _realutilization->update( make_pair( 0.0, time() ) );
        break;
    case server::RUNNING:
        _utilization->update( make_pair( 1.0, time() ) );
        _realutilization->update( make_pair( 1.0, time() ) );
        break;
    case server::BLOCKED:
        _utilization->update( make_pair( 1.0, time() ) );
        _realutilization->update( make_pair( 0.0, time() ) );
        break;
    default:
        cerr << "Unknown server status " << s << endl;
        exit( -1 );
    }
    _status = s;
}

void server::inner_body( void )
{
    while ( 1 ) {
        //
        // Phase 1: if the queue is empty, wait
        //
        if ( _jobs.empty( ) ) {
            setStatus( server::IDLE );
            passivate( );
	}
 
        //
        // Phase 2: queue is not empty. Process the first job
        //
        setStatus( server::RUNNING );        
        // dump_status();        
        job* j( _jobs.front( ) );  
        int jclass = j->getClass( );

        //
        // Phase 3: Serve the job
        //
        double servt = _servTime[jclass]->value();
        hold( servt );
        _completed++;

        //
        // Phase 4: Now, check if the destination server queue is not full
        //
        int nextServerNo = (int)_nextServer[ jclass ]->value();
        if ( -1 == nextServerNo ) {
            // Job exits the system
            delete j;
        } else {
            // Job is forwarded to another server
            assert( nextServerNo != _serverno );
            handle<server> next( servers[ nextServerNo ] );
            if ( next->full() ) {
                message( this, "Destination %d full. Will sleep\n", 
                         nextServerNo );
                // Queue full, sleep on the _blocked queue
                next->add_waiting_proc( current() );
                setStatus( server::BLOCKED );
                passivate( );
            } else {
                message( this, "Destination %d free. Will proceed\n", 
                         nextServerNo );
            }
            assert( !next->full() );
            next->add_new_job( j );
            // Wake up next server if idle
            if ( IDLE == next->getStatus() )
                next->activateAfter( current() );
        }    

        //
        // Phase 5: Remove the job from the queue
        //
        remove_first_job( );

        // Unblock the first server blocked on this node
        if ( !_blocked.empty() ) {
            handle_p f( _blocked.front( ) );
            _blocked.pop( );
            f->activateAfter( current() );
        }
    }
}
