/****************************************************************************
 *
 * cdf.hh
 *
 * This file is part of libcppsim.
 *
 * Copyright (C) 2005 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#ifndef _CPPSIM_CDF_HH
#define _CPPSIM_CDF_HH

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <string>
#include <cmath>

#ifndef squared
#define squared(x) ((x)*(x))
#endif

namespace cppsim {

    class cdf {
    public:
        virtual ~cdf( ) { };
        double m( void ) const { return _mean; };
        double v( void ) const { return _variance; };
        virtual double value( double x ) = 0;
    protected:
        cdf( const std::string& name, double m, double v ) : 
            _name( name ),
            _mean( m ),
            _variance( v )
        { };
        std::string _name;
        double _mean;
        double _variance;
    };

    class cdfUniform01 : public cdf {
    public:
        cdfUniform01( const std::string& name ) :
            cdf( name, .5, .25 ) { };
        double value( double );
    };

    class cdfUniform : public cdf {
    public:
        cdfUniform( const std::string& name, double a, double b ) :
            cdf( name, (a+b)/2.0, (b-a)*(b-a)/12.0 ),
            _a( a ),
            _b( b ) 
        { };
        double value( double );
    protected:
        double _a, _b;
    };

    class cdfExp : public cdf {
    public:
        cdfExp( const std::string& name, double m ) :
            cdf( name, m, m*m ) { };
        double value( double );
    };

    class cdfDiscUniform : public cdf {
    public:
        cdfDiscUniform( const std::string& name, int a, int b ) :
            cdf( name,(a+b)/2.0, ((b-a+1.0)*(b-a+1.0)-1.0)/12.0 ),
            _a( a ),
            _b( b )
        { };
        double value( double );
    protected:
        int _a, _b;
    };

    class cdfWeib : public cdf {
    public:
        cdfWeib( const std::string& name, double a, double b ) :
            cdf( name, a / b * tgamma( 1.0/b ),
                 squared(a) / squared(b) * (2*b*tgamma(2.0/b)-squared(tgamma(1.0/b)) ) ) { };
        double value( double );
    protected:
        double _a, _b;
    };

    class cdfErl : public cdf {
    public:
        cdfErl( const std::string& name, double m, int k ) :
            cdf( name, m, k ),
            _m( m ),
            _k( k )
        { };
        double value( double );
    protected:
        double _m;
        int _k;
    };

    class cdfBernoulli : public cdf {
    public:
        cdfBernoulli( const std::string& name, double p ) :
            cdf( name, p, p*(1.0-p) ),
            _p( p )
        { };
        double value( double );
    protected:
        double _p;
    };


    class cdfBinomial : public cdf {
    public:
        cdfBinomial( const std::string& name, int n, double p ) :
            cdf( name, n*p, n*p*(1.0-p) ),
            _n( n ),
            _p( p )
        { };
        double value( double );
    protected:
        int _n;
        double _p;
    };


    class cdfNormal : public cdf {
    public:
        cdfNormal( const std::string& name, double mu, double sigmasq ) :
            cdf( name, mu, sigmasq ),
            _mu( mu ),
            _sigmasq( sigmasq )
        { };
        double value( double );
    protected:
        double _mu;
        double _sigmasq;
    };


    class cdfChisq : public cdf {
    public:
        cdfChisq( const std::string& name, int n ):
            cdf( name, (double)n, 2.0*n ),
            _n( n )
        { };
        double value( double );
    protected:
        int _n;
    };


    class cdfGamma : public cdf {
    public:
        cdfGamma( const std::string& name, double a, double b ) :
            cdf( name, a*b, a*a*b ),
            _a( a ),
            _b( b )
        { };
        double value( double );
    protected:
        double _a, _b;
    };

};

#endif
