/*****************************************************************************
 *
 * mean.cc
 *
 * Compute the mean of a sequence of UNCORRELATED observations
 * x_1, x_2, \ldots x_N.
 * This file is part of libcppsim.
 *
 * Copyright (C) 2002, 2003, 2004 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <cmath>
#include <vector>
#include "mean.hh"
#include "statfun.hh"
#include "assertions.hh" // (r)
#include "oform.hh" // (r)

namespace cppsim {

    using namespace std;
    
    mean::mean( const string& name, double confl ) :
        statistic   ( name, confl ),
        _sum        ( 0.0 ),
        _sumsq      ( 0.0 )
    {
        
    }
    
    mean::~mean( )
    {
        
    }
    
    void mean::update( double v )
    {
        var_impl::update( v );
        _sum += v;
        _sumsq += squared(v);
    }
    
    void mean::reset( void )
    {
        var_impl::reset( );
        _sum = _sumsq = 0.0;
    }
    
    confInt mean::value( void ) throw( runtime_error )
    {
        //
        // Banks, p. 227
        //
        Assert( numUpdates() > 0,
                runtime_error( 
                              oform( "mean(\"%s\")::value(): "
                                     "Computing the mean with no observations",
                                     name().c_str() ) 
                              )
                );
        double w = qnorm( 1.0 - alpha()/2.0 )*sqrt( variance()/(double)numUpdates() );
        return confInt( m()-w, m()+w, confl() );
    }
    
    double mean::m( void ) const throw( runtime_error )
    {
        Assert( numUpdates() > 0,
                runtime_error( 
                              oform( "mean(\"%s\"::m(): "
                                     "Computing the mean with no observations",
                                     name().c_str() ) 
                              )
                );
        return _sum / (double)numUpdates();
    }
    
    double mean::variance( void ) const throw( runtime_error )
    {
        Assert( numUpdates() > 1,
                runtime_error( 
                              oform( "mean(\"%s\"::m(): "
                                     "Computing the variance with less than 2 observations",
                                     name().c_str() ) 
                              )
                );
        //
        // Note: fabs() is necessary as numerical errors may cause the
        // variance to be a tiny negative number. In this way the sqrt()
        // function on function mean() will fail and cause problems
        //
        return fabs( 
                    ( _sumsq - _sum*_sum / (double)numUpdates( ) ) / 
                    (double)( numUpdates( )-1 ) 
                    );
    }

    const vector< string > mean::headers( void )
    {
        vector< string > result(5);
        result[0] = "Name";
        result[1] = "Observations";
        result[2] = "Lower Bound";
        result[3] = "Upper Bound";
        result[4] = "Confidence Level";

        return result;
    }
    
    const vector< string > mean::entries( void )
    {
        vector< string > result(5);
        confInt c = const_cast<mean*>(this)->value( );

        result[0] = name();
        result[1] = oform("%d", numUpdates() );
        result[2] = oform("%f",c.lBound() );
        result[3] = oform("%f",c.uBound() );
        result[4] = oform("%f", confl() );

        return result;
    }
}
