/****************************************************************************
 *
 * quantile.cc
 *
 * Compute a confidence interval for the q-quantile of a sequence of
 * observations.  
 * This file is part of libcppsim.
 *
 * Copyright (C) 2003, 2004 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#include <algorithm>
#include <cmath>
#include "quantile.hh"
#include "statfun.hh"
#include "assertions.hh"

namespace cppsim {

    using namespace std;

    quantile::quantile( const string& name, double p, double confl ) :
        statistic( name, confl ),
        m_p( p )
    {
        Assert1( p > 0.0 );
        Assert1( p < 1.0 );
    }

    quantile::~quantile( )
    {

    }

    void quantile::update( double v )
    {
        m_numUpdates++;
        m_ob.push_back( v );
    }

    void quantile::reset( void )
    {
        m_numResets++;
        m_numUpdates = 0;
        m_ob.clear( );
    }

    //
    // Returns the value x such that P( X <= x ) = p
    //
    // FIXME: usare nth_element di STL?
    confInt quantile::value( void )
    {
        unsigned int k = m_ob.size();
        unsigned int i,j;

        // Sort the observations
        sort( m_ob.begin(), m_ob.end() );
    
        i = (unsigned int)ceil( k*m_p + 1.0 -
                                qnorm( 1.0 - alpha()/2.0 )*sqrt( k*m_p*(1.0-m_p) ) );
        j = (unsigned int)ceil( k*m_p + 1.0 +
                                qnorm( 1.0 - alpha()/2.0 )*sqrt( k*m_p*(1.0-m_p) ) );
        return confInt( m_ob[i], m_ob[j], confl() );
    }

    void quantile::report( void ) const
    {

    }

}
