/****************************************************************************
 *
 * repSimulation.cc
 *
 * Implementation of the independent replication methods for
 * steady-state simulation.
 *
 * This file is part of libcppsim.
 *
 * Copyright (C) 2004 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#include <functional>
#include <algorithm>
#include "repSimulation.hh"
#include "rng.hh"
#include "sqs.hh"
#include "statistic.hh"
#include "assertions.hh"        // (r)

namespace cppsim {

    using namespace std;

    void repSimulation::guard::inner_body( void )
    {
        hold( _duration );
        end_simulation( );
    }

    repSimulation::repSimulation(double simDuration,
                                 int maxReplications,
                                 double relWidth ) :
        simulation      ( ),
        _simDuration    ( simDuration ),
        _maxReplications( maxReplications ),
        _relWidth       ( relWidth ),
        _nrep           ( 0 )
    {

    }

    //
    // Prepare for the next execution of the simulation run
    //
    void repSimulation::reset( void )
    {
        for_each( _rngs.begin(), _rngs.end(), 
                  mem_fun_t< void, rng >( & rng::reset ) );
        for_each( _stats.begin(), _stats.end(),
                  mem_fun_t< void, statistic >( & statistic::reset ) );
    }

    //
    // Returns true iff the termination criteria is satisfied. The
    // stopping criteria is: (a) the number of replications exceeds
    // the maximum allowed value, OR (b) all the statistics in the
    // _stats array converged within the specified confidence interval
    // relative width.
    //
    bool repSimulation::terminated( void )
    {
        if ( _nrep >= _maxReplications )
            return true;

        list<statistic*>::iterator it;
        try {
            for ( it=_stats.begin(); it!=_stats.end(); it++ ) {
                confInt c( (*it)->value() );
                if ( (c.uBound() - c.lBound())/c.val() > _relWidth )
                    return false;
            }
        } catch( runtime_error& e ) {
            return false;
        }
        return true;
    }

    //
    // Executes the simulation
    //
    void repSimulation::run( void ) 
    {
        Assert1( ! _sim.empty() );
        Assert1( this == _sim.front() );
        bool t = false;
        cout << "Run#\tTerminated?" << endl << endl;
        do {
            _nrep++;
            cout << _nrep << '\t' << flush;
            handle_p g( new guard( _simDuration ) );
            g->activate( );
            initialize( );
            _sqs->call( );
            finalize( );
            t = terminated( );
            if ( !t ) {
                cout << 'N' << endl;
                reset( );
                // Reinitialize the SQS
                delete _sqs; _sqs = defaultSqsBuilder();
            }
        } while ( !t );
        cout << 'Y' << endl;
    }

}
