/*****************************************************************************
 *
 * rng.cc
 *
 * Random Number Generators classes.
 * This file is part of libcppsim.
 *
 * Copyright (C) 2005 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <cmath>
#include <iostream>
#include <string>
#include <cstdio>
#include <cstdlib>
// #include <numeric>
#include "rng.hh"
#include "cdf.hh"
#include "assertions.hh" // (r)
#include "oform.hh" // (r)

namespace cppsim {

    using namespace std;

    //////////////////////////////////////////////////////////////////////////
    //
    // rng
    //
    //////////////////////////////////////////////////////////////////////////
    rng::rng( const string& name ) :
        var_impl        ( name, reportable::RNG ),
        _numDraws       ( 0 )
    { 
        setClass( reportable::RNG );
    }; 

    string rng::title( void ) const 
    {
        return string("Random Number Generators" );
    };

    const vector< string > rng::headers( void )
    {
        vector< string > result(3);
        result[0] = "Name";
        result[1] = "Draws";
        result[2] = "Resets";
        return result;        
    }

    const vector< string > rng::entries( void ) 
    {
        vector< string > result(3);
        result[0] = name().c_str( );
        result[1] = oform("%d", _numDraws );
        result[2] = oform("%d", numResets() );        
        return result;
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // rngUniform01
    //
    //////////////////////////////////////////////////////////////////////////

    rngUniform01::seed_t rngUniform01::_globalSeed = { 0.0, 0.0, 1.0, 0.0, 0.0, 1.0 };
    const double rngUniform01::norm  = 2.328306549295728e-10;  /* 1.0/(m1+1) */
    const double rngUniform01::norm2 = 2.328318825240738e-10;  /* 1.0/(m2+1) */
    const double rngUniform01::m1 = 4294967087.0;
    const double rngUniform01::m2 = 4294944443.0;
    unsigned int rngUniform01::maxSeqLen = 1000;

    double rngUniform01::draw( seed_t s )
    {
        long k;
        double p,
            s10 = s[0], s11 = s[1], s12 = s[2],
            s20 = s[3], s21 = s[4], s22 = s[5];

        p = 1403580.0 * s11 - 810728.0 * s10;
        k = (long)(p / m1);  p -= k*m1;  if (p < 0.0) p += m1;
        s10 = s11;   s11 = s12;  s12 = p;

        p = 527612.0 * s22 - 1370589.0 * s20;
        k = (long)(p / m2);  p -= k*m2;  if (p < 0.0) p += m2;
        s20 = s21;   s21 = s22;  s22 = p; 

        s[0] = s10;  s[1] = s11;  s[2] = s12;
        s[3] = s20;  s[4] = s21;  s[5] = s22;

        if (s12 <= s22) 
            return ((s12 - s22 + m1) * norm);
        else 
            return ((s12 - s22) * norm);
    }

    //
    // Advance the global seed space
    //
    void rngUniform01::advance( unsigned int count ) 
    {
        while( count-- > 0 )
            draw( _globalSeed );
    }

    rngUniform01::rngUniform01( const string& name ) :
        rng( name ),
        _seqLen( maxSeqLen )
    {
        memcpy( _mySeed, _globalSeed, sizeof( _globalSeed ) );
        advance( maxSeqLen );
    }

    double rngUniform01::value( void )
    {
        double result = draw( _mySeed );
        if ( 0 == _seqLen-- ) {
            _seqLen = maxSeqLen;
            memcpy( _mySeed, _globalSeed, sizeof( _globalSeed ) );
            advance( maxSeqLen );
        }
        return result;
    }

    void rngUniform01::reset( void )
    {
        rng::reset();
        memcpy( _mySeed, _globalSeed, sizeof( _globalSeed ) );
        _seqLen = maxSeqLen;
        advance( maxSeqLen );
    }

    cdf* rngUniform01::makeCdf( void ) const {
        return new cdfUniform01( "cdf::"+name() );
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // Uniform distribution over [a,b]
    //
    //////////////////////////////////////////////////////////////////////////

    rngUniform::rngUniform( const string& name, double a, double b ) 
        throw( domain_error ) :
        rng ( name ),
        _a ( a ),
        _b ( b ),
        U1 ( "U1_"+name )
    {
        Assert( _b > _a,
                domain_error( oform( "rngUniform(\"%s\",a=%f, b=%f) must have "
                                     "b>a", name.c_str(), a, b ) 
                              )
                );
    }

    double rngUniform::value( void )
    {
        _numDraws++;
        return U1.value( )*(_b - _a) + _a;
    }

    void rngUniform::reset( void )
    {
        rng::reset( );
        U1.reset( );
    }

    cdf* rngUniform::makeCdf( void ) const
    {
        return new cdfUniform( "cdf::"+name(), _a, _b );
    }

    //////////////////////////////////////////////////////////////////////////
    //
    // Discrete Uniform distribution over {a, a+1, ..., b-1, b}
    //
    //////////////////////////////////////////////////////////////////////////

    rngDiscUniform::rngDiscUniform( const string& name, int a, int b ) 
        throw( domain_error ) :
        rng ( name ),
        _a ( a ),
        _b ( b ),
        U1 ( "U1_"+name )
    {
        Assert( _b >= _a,
                domain_error( oform( "rngDiscUniform(\"%s\",a=%d, b=%d) must "
                                     "have b>=a", name.c_str(), a, b ) 
                              )
                );
    }

    double rngDiscUniform::value( void )
    {
        _numDraws++;
        const double _aa = (double)_a - 0.499999999;
        const double _bb = (double)_b + 0.499999999;
        int result = (int)( rint( U1.value( )*(_bb - _aa) + _aa ) );
        Assert1( result >= _a );
        Assert1( result <= _b );
        return (double)result;
    }

    void rngDiscUniform::reset( void )
    {
        rng::reset();
        U1.reset( );
    }

    cdf* rngDiscUniform::makeCdf( void ) const
    {
        return new cdfDiscUniform( "cdf::"+name(), _a, _b );
    }


    //////////////////////////////////////////////////////////////////////////
    //
    // Discrete empirical distribution. Returns the integers {0, 1, ...n}
    // with probability p[0], p[1], ... p[n].
    // If p[0]+p[1]+...+p[n] is less than 1.0, it *may* return -1 to denote
    // that none has been choose.
    //
    //////////////////////////////////////////////////////////////////////////
    rngDiscEmpirical::rngDiscEmpirical( const string& name, const vector<double>& p ) 
        throw( domain_error ) :
        rng ( name ),
        _p ( p ),
        //        _psum ( p.size(), 0 ),
        U1 ( "U1_"+name )
    {
        //        partial_sum( _p.begin(), _p.end(), _psum.begin() );
        Assert( p.size() > 0,
                domain_error( oform( "rngDiscEmpirical(\"%s\",p[]) vector "
                                     "p is empty", name.c_str() ) 
                              )
                );
    }

    double rngDiscEmpirical::value( void )
    {
        _numDraws++;
        const double tmp = U1.value( );
        double sum = 0.0;
        int i;
        for ( i=0; i<(int)_p.size( ); i++ ) {
            sum += _p[i];
            if ( tmp <= sum ) return (double)i;
        }
        return (double)-1;
//         _numDraws++;
//         const double tmp = U1.value( );
//         for ( int i=0; i<(int)_psum.size(); i++ ) {
//             if ( tmp <= _psum[i] )
//                 return (double)i;
//         }
//         return (double)-1;
    }

    void rngDiscEmpirical::reset( void )
    {
        rng::reset();
        U1.reset( );
    }


    //////////////////////////////////////////////////////////////////////////
    //
    // Exponential Distribution
    //
    //////////////////////////////////////////////////////////////////////////
    rngExp::rngExp( const string& name, double mean )
        throw( domain_error ) :
        rng ( name ),
        _mean ( mean ),
        U1 ( "U1_"+name )
    {
        Assert( _mean > 0.0,
                domain_error( oform( "rngExp(\"%s\", mean=%f): must be mean>0",
                                     name.c_str(), _mean )
                              )
                );
    }

    double rngExp::value( void )
    {
        _numDraws++;
        return (double)(-log( 1.0 - U1.value( ) ) * _mean );
    }

    void rngExp::reset( void )
    {
        rng::reset();
        U1.reset( );
    }

    cdf* rngExp::makeCdf( void ) const
    {
        return new cdfExp( "cdf::"+name(), _mean );
    }

    //////////////////////////////////////////////////////////////////////////
    //
    // Weibull distribution
    //
    //////////////////////////////////////////////////////////////////////////
    rngWeib::rngWeib( const string& name, double a, double b ) 
        throw( domain_error ) :
        rng         ( name ),
        _a          ( a ),
        _b          ( b ),
        U1          ( "U1_"+name )
    {
        Assert( a > 0.0 && b > 0.0,
                domain_error( oform( "rngWeib(\"%s\", a=%f,b=%f): must be a>0 and b>0",
                                     name.c_str(), a, b )
                              )
                );
    };

    double rngWeib::value( void )
    {
        _numDraws++;
        return _a*pow( -log( 1.0 - U1.value() ), 1.0/_b );
    };

    void rngWeib::reset( void )
    {
        rng::reset();
        U1.reset( );
    }

    cdf* rngWeib::makeCdf( void ) const
    {
        return new cdfWeib( "cdf::"+name(), _a, _b );
    }

    //////////////////////////////////////////////////////////////////////////
    //
    // Erlang
    //
    //////////////////////////////////////////////////////////////////////////
    rngErl::rngErl( const string& name, double m, int k ) 
        throw( domain_error ) :
        rng         ( name ),
        _m          ( m ),
        _k          ( k ),
        _rng        ( 0 )
    {
        Assert( m > 0.0 && k > 0,
                domain_error( oform( "rngErl(\"%s\", m=%f,k=%d): must be m>0 and k>0",
                                     name.c_str(), m, k )
                              )
                );
        if ( _k < 10 )
            _rng = new rngUniform01( "U1_"+name );
        else
            _rng = new rngGamma( "GAM_"+name, m/(double)k, k );
    }

    rngErl::~rngErl( )
    {
        delete _rng;
    }

    double rngErl::value( void )
    {
        _numDraws++;
        if ( _k < 10 ) {
            double prod = 1.0;
            for ( int i=0; i<_k; i++ )
                prod *= ( 1.0 - _rng->value() );
            return -( _m/(double)_k )*log( prod );
        } else {
            return _rng->value();
        }
    }

    void rngErl::reset( void )
    {
        rng::reset();
        _rng->reset( );
    }

    cdf* rngErl::makeCdf( void ) const
    {
        return new cdfErl( "cdf::"+name(), _m, _k );
    }

    //////////////////////////////////////////////////////////////////////////
    //
    // Bernoulli distribution (returns true with probability p, false with pro-
    // bability 1-p),
    //
    //////////////////////////////////////////////////////////////////////////
    rngBernoulli::rngBernoulli( const string& name, double p )
        throw( domain_error ) :
        rng         ( name ),
        _p          ( p ),
        U1          ( "U01_"+name )
    {
        Assert( p >= 0.0 && p <= 1.0,
                domain_error( oform( "rngBernoulli(\"%s\", p=%f): must be 0<=p<=1",
                                     name.c_str(), p )
                              )
                );
    }

    double rngBernoulli::value( void )
    {
        _numDraws++;
        return (double)( U1.value( ) <= _p );
    }

    void rngBernoulli::reset( void )
    {
        rng::reset();
        U1.reset( );
    }

    cdf* rngBernoulli::makeCdf( void ) const
    {
        return new cdfBernoulli( "cdf::"+name(), _p );
    }

    //////////////////////////////////////////////////////////////////////////
    //
    // Binomial distribution 
    //
    //////////////////////////////////////////////////////////////////////////
    rngBinomial::rngBinomial( const string& name, int n, double p ) 
        throw( domain_error ) :
        rng         ( name ),
        _n          ( n ),
        _p          ( p ),
        U1          ( "U01_"+name, p )
    {
        Assert( n > 0,
                domain_error( oform( "rngBinomial(\"%s\", n=%d, p=%f): must be n>0",
                                     name.c_str(), n, p )
                              )
                );
    }

    double rngBinomial::value( void )
    {
        _numDraws++;
        double result = 0.0;
        for ( int i=0; i<_n; i++ )
            result += U1.value();
        return result;
    }

    void rngBinomial::reset( void )
    {
        rng::reset();
        U1.reset( );
    }

    cdf* rngBinomial::makeCdf( void ) const
    {
        return new cdfBinomial( "cdf::"+name(), _n, _p );
    }


    //////////////////////////////////////////////////////////////////////////
    //
    // Normal distribution N( mu, sigmasq )
    //
    //////////////////////////////////////////////////////////////////////////
    rngNormal::rngNormal( const string& name, double mu, double sigmasq )
        throw( domain_error ) :
        rng         ( name ),
        _mu         ( mu ),
        _sigmasq    ( sigmasq ),
        U1          ( "U1_"+name ),
        U2          ( "U2_"+name )
    {
        Assert( sigmasq > 0.0,
                domain_error( oform( "rngNormal(\"%s\", mu=%f, sigmasq=%f): must be sigmasq>0",
                                     name.c_str(), mu, sigmasq )
                              )
                );
        val[0] = val[1] = 0.0;
    }

    // Correction with respect to Banks. See Jain, p. 494
    // Generated using the Box-Muller method.
    double rngNormal::value( void )
    {
        if ( 0 == ( _numDraws % 2 ) ) {
            double R = sqrt( -2.0*log( U1.value( ) ) );
            double T = 2.0*M_PI*U2.value( );
            val[0] = _mu + sqrt( _sigmasq )*R*cos( T );
            val[1] = _mu + sqrt( _sigmasq )*R*sin( T );
        }
        return val[ _numDraws++ % 2 ];
    }

    void rngNormal::reset( void )
    {
        rng::reset();
        U1.reset( );
        U2.reset( );
    }

    cdf* rngNormal::makeCdf( void ) const
    {
        return new cdfNormal( "cdf::"+name(), _mu, _sigmasq );
    }


    //////////////////////////////////////////////////////////////////////////
    //
    // Left truncated Normal distribution N( mu, sigmasq, l )
    //
    //////////////////////////////////////////////////////////////////////////
    rngLTNormal::rngLTNormal( const string& name, double mu, double sigmasq, double l )
        throw( domain_error ) :
        rng         ( name ),
        _mu         ( mu ),
        _sigmasq    ( sigmasq ),
        _l          ( l ),
        U1          ( "U1_"+name, mu, sigmasq )
    {
        Assert( sigmasq > 0.0,
                domain_error( oform( "rngLTNormal(\"%s\", mu=%f, sigmasq=%f): must be sigmasq>0",
                                     name.c_str(), mu, sigmasq )
                              )
                );
    }

    double rngLTNormal::value( void )
    {
        double result;
        _numDraws++;
        do {
            result = U1.value( );
        } while ( result < _l );
        return result;
    }

    void rngLTNormal::reset( void )
    {
        rng::reset();
        U1.reset( );
    }


    //////////////////////////////////////////////////////////////////////////
    //
    // Right truncated Normal distribution N( mu, sigmasq, r )
    //
    //////////////////////////////////////////////////////////////////////////
    rngRTNormal::rngRTNormal( const string& name, double mu, double sigmasq, double r )
        throw( domain_error ) :
        rng         ( name ),
        _mu         ( mu ),
        _sigmasq    ( sigmasq ),
        _r          ( r ),
        U1          ( "U1_"+name, mu, sigmasq )
    {
        Assert( sigmasq > 0.0,
                domain_error( oform( "rngRTNormal(\"%s\", mu=%f, sigmasq=%f): must be sigmasq>0",
                                     name.c_str(), mu, sigmasq )
                              )
                );
    }

    double rngRTNormal::value( void )
    {
        double result;
        _numDraws++;
        do {
            result = U1.value( );
        } while ( result > _r );
        return result;
    }

    void rngRTNormal::reset( void )
    {
        rng::reset();
        U1.reset( );
    }


    //////////////////////////////////////////////////////////////////////////
    //
    // Left and Right truncated Normal distribution N( mu, sigmasq, l, r )
    //
    //////////////////////////////////////////////////////////////////////////
    rngLRTNormal::rngLRTNormal( const string& name, double mu, double sigmasq, double l, double r )
        throw( domain_error ) :
        rng         ( name ),
        _mu         ( mu ),
        _sigmasq    ( sigmasq ),
        _l          ( l ),
        _r          ( r ),
        U1          ( "U1_"+name, mu, sigmasq )
    {
        Assert( sigmasq > 0.0,
                domain_error( oform( "rngRLTNormal(\"%s\", mu=%f, sigmasq=%f): must be sigmasq>0",
                                     name.c_str(), mu, sigmasq )
                              )
                );
    }

    double rngLRTNormal::value( void )
    {
        double result;
        _numDraws++;
        do {
            result = U1.value( );
        } while ( ( result < _l ) || ( result > _r ) );
        return result;
    }

    void rngLRTNormal::reset( void )
    {
        rng::reset();
        U1.reset( );
    }


    //////////////////////////////////////////////////////////////////////////
    //
    // Chi-square distribution with n degrees of freedom
    //
    //////////////////////////////////////////////////////////////////////////
    rngChisq::rngChisq( const string& name, int n ) 
        throw( domain_error ) :
        rng         ( name ),
        _n          ( n ),
        U1          ( "U1_"+name, 2.0, (double)n/2.0 )
    {
        Assert( _n > 0,
                domain_error( oform( "rngChisq(\"%s\", n=%d): n must be >0",
                                     name.c_str(), _n )
                              )
                );
    }

    double rngChisq::value( void )
    {
        _numDraws++;
        return U1.value( );
    }

    void rngChisq::reset( void )
    {
        rng::reset( );
        U1.reset( );
    }

    cdf* rngChisq::makeCdf( void ) const
    {
        return new cdfChisq( "cdf::"+name(), _n );
    }

    //////////////////////////////////////////////////////////////////////////
    //
    // Gamma(a,b) distribution. Banks, p. 152
    //
    //////////////////////////////////////////////////////////////////////////
    rngGamma::rngGamma( const string& name, double a, double b ) 
        throw( domain_error ) :
        rng         ( name ),
        _a          ( a ),
        _b          ( b ),
        U1          ( "U1_"+name ),
        U2          ( "U2_"+name )
    {
        Assert( a > 0.0 && b > 0.0 && b != 1.0,
                domain_error( oform( "rngGamma(\"%s\", a=%f,b=%f): must be a>0 and b>0, b != 1",
                                     name.c_str(), a, b )
                              )
                );
    }

    double rngGamma::value( void )
    {
        _numDraws++;
        if ( _b < 1.0 ) {
            double beta = ( M_E + _b ) / M_E;
            while ( 1 ) {
                double U = U1.value();
                double W = beta*U;
                if ( W < 1.0 ) {
                    double Y = pow( W, 1.0/_b );
                    double V = U2.value();
                    if ( V <= exp( -Y ) )
                        return _a*Y;
                } else {
                    double Y = - log( ( beta - W ) / _b );
                    double V = U2.value();
                    if ( V <= pow( Y, _b-1.0 ) )
                        return _a*Y;
                }
            }
        } else { // b > 1.0
            double alfa = 1.0 / sqrt( 2*_b-1 );
            double beta = _b - log( 4.0 );
            double gamma = _b + 1.0 / alfa;
            double delta = 1.0 + log( 4.5 );
            while ( 1 ) {
                double u1 = U1.value();
                double u2 = U2.value();
                double V = alfa*log( u1/(1.0-u1) );
                double Y = _b*exp( V );
                double Z = u1*u1*u2;
                double W = beta + gamma*V - Y;
                if ( W+delta-4.5*Z >= 0.0 )
                    return _a*Y;
                else {
                    if ( W >= log( Z ) )
                        return _a*Y;
                }
            }
        }
    }

    void rngGamma::reset( void )
    {
        rng::reset( );
        U1.reset( );
        U2.reset( );
    }

    cdf* rngGamma::makeCdf( void ) const
    {
        return new cdfGamma( "cdf::"+name(), _a, _b );
    }

}
