/*****************************************************************************
 *
 * rng.hh
 *
 * Random Number Generators classes.
 * This file is part of libcppsim. 
 *
 * Copyright (C) 2002, 2003, 2004 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#ifndef _CPPSIM_RNG_HH
#define _CPPSIM_RNG_HH

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <string>
#include <vector>
#include <iostream>
#include <stdexcept>
#include "var.hh"

#ifndef squared
#define squared(x) ((x)*(x))
#endif

namespace cppsim {

    class cdf;

    class rng : public var< double, double > {
    public:
        typedef var< double, double > var_impl;
        virtual ~rng( ) { };
        // Accessors
        string title( void ) const;
        const vector< string > headers( void );
        const vector< string > entries( void );
        unsigned int numDraws( void ) const { return _numDraws; };
        virtual cdf* makeCdf( void ) const { return 0; };
    protected:
        rng( const std::string& name );

        unsigned int _numDraws;
    private:
        void update( double v ) // This method is disabled for rngs
        { 
            std::cerr << "FATAL: rng::update( ) called for " << name() << std::endl;
            abort( );
        };
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // Uniform distribution in [0,1]
    //
    //////////////////////////////////////////////////////////////////////////
    class rngUniform01 : public rng {
    public:
        rngUniform01( const std::string& name );
        virtual ~rngUniform01( ) { };
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;

        static unsigned int maxSeqLen;

    protected:

        typedef double seed_t[6];

        double draw( seed_t );
        void advance( unsigned int count );

        seed_t _mySeed;
        unsigned int _seqLen;

        // Static members
        static seed_t _globalSeed;
        static const double norm;
        static const double norm2;
        static const double m1;
        static const double m2;
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // Uniform distribution in [a,b]
    //
    // Density function: f(x) = 1/(b-a) if a < x < b
    //                        = 0       otherwise
    //
    //////////////////////////////////////////////////////////////////////////
    class rngUniform : public rng {
    public:
        rngUniform( const std::string& name, double a, double b ) 
            throw( std::domain_error );
        virtual ~rngUniform( ) { };
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;
    protected:
        double      _a, _b;
        rngUniform01 U1;
    };

    //////////////////////////////////////////////////////////////////////////
    // 
    // Weibull distribution
    //
    // Density Function: f(x) = ba^(-b )x^(b-1)exp( -(x/a)^b ) if x > 0
    //                          0                              otherwise
    //
    //////////////////////////////////////////////////////////////////////////
    class rngWeib : public rng {
    public:
        rngWeib( const std::string& name, double a, double b ) 
            throw( std::domain_error );
        virtual ~rngWeib( ) { };
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;
    protected:
        double      _a, _b;
        rngUniform01 U1;
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // Constant distribution (always returns mean)
    //
    //////////////////////////////////////////////////////////////////////////
    class rngConst : public rng {
    public:
        rngConst( const std::string& name, double mean ) : 
            rng( name ),
            _mean( mean ) 
        { };
        virtual ~rngConst( ) { };
        double value( void ) { 
            _numDraws++; 
            return _mean; 
        };
        void reset( void ) { 
            m_numUpdates = 0;
            m_numResets++;
        };
    protected:
        double      _mean;
    };


    //////////////////////////////////////////////////////////////////////////
    //
    // Bernoulli
    //
    //////////////////////////////////////////////////////////////////////////
    class rngBernoulli : public rng {
    public:
        rngBernoulli( const std::string& name, double p )
            throw( std::domain_error );
        virtual ~rngBernoulli( ) { };
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;
    protected:
        double      _p;
        rngUniform01 U1;
    };


    //////////////////////////////////////////////////////////////////////////
    //
    // Binomial
    //
    //////////////////////////////////////////////////////////////////////////
    class rngBinomial : public rng {
    public:
        rngBinomial( const std::string& name, int n, double p )
            throw( std::domain_error );
        virtual ~rngBinomial( ) { };
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;
    protected:
        int     _n;
        double  _p;
        rngBernoulli U1;
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // k-Erlang
    //
    //////////////////////////////////////////////////////////////////////////
    class rngErl : public rng {
    public:
        rngErl( const std::string& name, double m, int k )
            throw( std::domain_error );
        virtual ~rngErl( );
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;
    protected:
        double      _m;
        int         _k;
        rng        *_rng;
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // Normal
    //
    //////////////////////////////////////////////////////////////////////////
    class rngNormal : public rng {
    public:
        rngNormal( const std::string& name, double mu, double sigmasq )
            throw( std::domain_error );
        virtual ~rngNormal( ) { };
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;
    protected:
        double      _mu;
        double      _sigmasq;
        rngUniform01 U1;
        rngUniform01 U2;
        double      val[2];
    };


    //////////////////////////////////////////////////////////////////////////
    //
    // Left-truncated normal distribution
    //
    //////////////////////////////////////////////////////////////////////////
    class rngLTNormal : public rng {
    public:
        rngLTNormal( const std::string& name, double mu, double sigmasq, double l )
            throw( std::domain_error );
        virtual ~rngLTNormal( ) { };
        double value( void );
        void reset( void );
    protected:
        double      _mu;
        double      _sigmasq;
        double      _l;
        rngNormal   U1;
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // Right-truncated normal distribution
    //
    //////////////////////////////////////////////////////////////////////////
    class rngRTNormal : public rng {
    public:
        rngRTNormal( const std::string& name, double mu, double sigmasq, double r )
            throw( std::domain_error );
        virtual ~rngRTNormal( ) { };
        double value( void );
        void reset( void );
    protected:
        double      _mu;
        double      _sigmasq;
        double      _r;
        rngNormal   U1;
    };


    //////////////////////////////////////////////////////////////////////////
    //
    // Left- and Right-truncated normal distribution
    //
    //////////////////////////////////////////////////////////////////////////
    class rngLRTNormal : public rng {
    public:
        rngLRTNormal( const std::string& name, double mu, double sigmasq, double l, double r )
            throw( std::domain_error );
        virtual ~rngLRTNormal( ) { };
        double value( void );
        void reset( void );
    protected:
        double      _mu;
        double      _sigmasq;
        double      _l;
        double      _r;
        rngNormal   U1;
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // Discrete Uniform distribution in {a, a+1, ..., b-1, b}. Equivalent
    // to the constant distribution if a==b.
    //
    //////////////////////////////////////////////////////////////////////////
    class rngDiscUniform : public rng {
    public:
        rngDiscUniform( const std::string& name, int a, int b )
            throw( std::domain_error );
        virtual ~rngDiscUniform( ) { };
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;
    protected:
        int _a, _b;
        rngUniform01 U1;
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // Discrete Empirical Distribution
    //
    //////////////////////////////////////////////////////////////////////////
    class rngDiscEmpirical : public rng {
    public:
        rngDiscEmpirical( const std::string& name, const std::vector<double>& p )
            throw( std::domain_error );
        virtual ~rngDiscEmpirical( ) { };
        double value( void );
        void reset( void );
    protected:
        std::vector<double> _p;
        // std::vector<double> _psum;
        rngUniform01 U1;
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // Exponential distribution
    //
    // Density function: f(x) = a exp( -ax ) if x > 0
    //                        = 0            otherwise
    //
    //////////////////////////////////////////////////////////////////////////
    class rngExp : public rng {
    public:
        rngExp( const std::string& name, double mean ) 
            throw( std::domain_error ) ;
        virtual ~rngExp( ) { };
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;
    protected:
        double      _mean;
        rngUniform01 U1;
    };

    //////////////////////////////////////////////////////////////////////////
    //
    // Gamma(a,b) distribution
    //
    //////////////////////////////////////////////////////////////////////////
    class rngGamma : public rng {
    public:
        rngGamma( const std::string& name, double a, double b )
            throw( std::domain_error );
        virtual ~rngGamma( ) { };
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;
    protected:
        double _a, _b;
        rngUniform01 U1, U2;
    };


    //////////////////////////////////////////////////////////////////////////
    //
    // Chi square distribution with n degrees of freedom
    //
    //////////////////////////////////////////////////////////////////////////
    class rngChisq : public rng {
    public:
        rngChisq( const std::string& name, int n )
            throw( std::domain_error );
        virtual ~rngChisq( ) { };
        double value( void );
        void reset( void );
        cdf* makeCdf( void ) const;
    protected:
        int _n;
        rngGamma U1;
    };

}

#endif
