/****************************************************************************
 *
 * timeweighted.cc
 *
 * This file is part of libcppsim.
 *
 * Copyright (C) 2004 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#include "timeweighted.hh"
#include "repmean.hh"
#include "assertions.hh" // (r)
#include "welch.hh"

namespace cppsim {

    using namespace std;

    timeweighted::timeweighted( const string& name, double confl, 
                                double bwidth, trremoval* t ) :
        var_impl( name, reportable::STATISTIC ),
        _confl( confl ),
        _bwidth( bwidth ),
        _startTime( 0 ),
        _lastUpdated( 0 ),
        _lastVal( 0 ),
        _sum( 0 ),
        _r( "repmean("+name+")::_r", confl, t )
        // _w( new welch( "welch_"+name+"_r", confl, 5 ) )
    {

    }


    timeweighted::~timeweighted( )
    {
        // delete _w;
    }


    //
    // ----|----------------|-----------|--------->
    // _startTime       _startTime      t
    //                   +_bwidth
    //
    void timeweighted::update( const pair< double, double>& v )
    {
        double t = v.second;
        double val = v.first;
        Assert1( t >= _lastUpdated );
        Assert1( _lastUpdated >= _startTime );
        while ( t - _startTime > _bwidth ) {
            // Complete current batch
            _sum += ( _startTime + _bwidth - _lastUpdated ) * _lastVal;
            _r.update( _sum / _bwidth );
            // _w->update( _sum / _bwidth );
            // Start a new batch
            _startTime += _bwidth;
            _lastUpdated = _startTime;
            _sum = 0;
        }
        _sum += (t - _lastUpdated) * _lastVal;
        _lastVal = val;
        _lastUpdated = t;
        var_impl::update( v );
    }


    void timeweighted::reset( void )
    {
        // FIXME: se non sono state fatte operazioni di update() da
        // tempo, questa variabile potrebbe contenere meno osservazioni
        // di quelle effettivamente presenti
        var_impl::reset( );
        _r.reset( );
        // _w->reset( );
        _sum = _startTime = _lastUpdated = _lastVal = 0;
    }


    confInt timeweighted::value( void )
    {
        // _w->entries( );
        return _r.value( );
    }

}
