/****************************************************************************
 *
 * timewhisto.cc
 *
 * This file is part of libcppsim.
 *
 * Copyright (C) 2005 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#include "timewhisto.hh"
#include "timeweighted.hh"
#include "trremoval.hh"
#include "oform.hh" // (r)
#include "assertions.hh" // (r)
#include <string>

namespace cppsim {

    using namespace std;

    timewhisto::timewhisto( const string& name, int nmax, double confl, 
                            double bwidth ) :
        var_impl( name, reportable::STATISTIC ),
        _nmax( nmax ),
        _confl( confl ),
        _bins( nmax )
    {
        for ( int i=0; i<nmax; i++ ) {
            _bins[i] = new timeweighted( oform("%s_bin_%d", name.c_str(), i),
                                         confl, bwidth );
        }
    }


    timewhisto::~timewhisto( )
    {
        for ( int i=0; i<_nmax; i++ )
            delete _bins[i];
    }

    void timewhisto::update( const pair< int, double >& v )
    {
        var_impl::update( v );
        Assert1( v.first >= 0 );
        for ( int i=0; i<_nmax; i++ ) {
            _bins[i]->update( make_pair(
                                        ( i == v.first ? 1.0 : 0.0 ),
                                        v.second ) 
                              );                     
        }
    }


    void timewhisto::reset( void )
    {
        var_impl::reset( );
        for_each( _bins.begin(), _bins.end(), 
                  mem_fun( &timeweighted::reset ) );
    }


    confInt timewhisto::value( void )
    {
        return confInt( 0, 0, 0 );

        /*
          list< confp > result;
          for ( int i=0; i<_nmax; i++ ) 
          result.push_back( _bins[i]->value() );
          return result;
        */
    }

    const vector< string > timewhisto::headers( void )
    {
        vector< string > result(_nmax);
        for ( int i=0; i<_nmax; i++ ) 
            result[i] = oform("%s[%d]", name().c_str(), i );
        return result;
    }

    const vector< string > timewhisto::entries( void )
    {
        vector< string > result(_nmax);
        for ( int i=0; i<_nmax; i++ ) {
            confInt c( _bins[i]->value() );
            result[i] = oform("%f %f", c.lBound(), c.uBound() );
        }
        return result;        
    }

}
