/****************************************************************************
 *
 * trremoval.hh
 *
 * Transient Removal classes
 * This file is part of libcppsim
 *
 * Copyright (C) 2003, 2004 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/
#ifndef _CPPSIM_TRREMOVAL_HH
#define _CPPSIM_TRREMOVAL_HH

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <vector>
#include <functional>
#include "var.hh"
#include "assertions.hh"        // (r)

namespace cppsim {

    class trremoval : public var< double, std::vector<double> > {
    public:
        typedef var< double, std::vector<double> > var_impl;
        virtual ~trremoval( ) { };
        const std::vector< std::string > headers( void );
    protected:
        trremoval( const std::string& name ) :
            var_impl( name, reportable::TRREMOVAL )
        { };
    };

    class trremoval_const : public trremoval {
    public:
        trremoval_const( const std::string& name, const size_t l=0 ) :
            trremoval( name ),
            _l( l ) 
        { 

        };
        virtual ~trremoval_const( ) 
        { 

        };
        // Modifiers
        void update( double v ); 
        void reset( void );
        // Accessors
        std::vector<double> value( void ) throw( std::runtime_error )
        {
            return _ob;
        }
        const std::vector< string > entries( void );
    protected:
        size_t _l;                  // length of the warm-up period
        std::vector<double> _ob;    // list of (truncated) observations
    };

    class trremoval_frac : public trremoval {
    public:
        trremoval_frac( const std::string& name, float f ) :
            trremoval( name ),
            _f( f ) 
        { 
            Assert( 0.0 < f && f < 1.0,
                    std::domain_error( "Domain Error" ) );
        };
        virtual ~trremoval_frac( ) 
        { 

        };
        // Modifiers
        void update( double v );
        void reset( void );
        vector<double> value( void ) throw( std::runtime_error );
        const vector< string > entries( void );
    protected:
        float _f;       // Fraction of the observation list to remove
                        // e.g., 0.2 = drop first 20% of the list
        std::vector<double> _ob;        // list of observations
    };



    //
    // trremoval factory
    // Class used to instantiate a new trremoval object
    //
    class trremovalBuilder_t {
    public:
        virtual ~trremovalBuilder_t( ) { };
        virtual trremoval* operator()( void ) = 0;
    protected:
        trremovalBuilder_t( ) { };
    };

    // Creates a new trremoval_const object
    class trremovalBuilderConst : public trremovalBuilder_t {
    public:
        trremovalBuilderConst( unsigned int l ) : 
            trremovalBuilder_t( ),
            _l( l ) { };
        trremoval* operator()( void ) {
            return new trremoval_const( "trremovalBuilderConst", _l );
        };
    protected:
        unsigned int _l;
    };

    // Creates a new trremoval_frac object
    class trremovalBuilderFrac : public trremovalBuilder_t {
    public:
        trremovalBuilderFrac( float f ) : 
            trremovalBuilder_t( ),
            _f( f ) { };
        trremoval* operator()( void ) {
            return new trremoval_frac( "trremovalBuilderFrac", _f );
        };
    protected:
        float _f;
    };

    extern trremovalBuilder_t *trremovalDefaultBuilder;


}

#endif
