/****************************************************************************
 *
 * mm1_test_2.cc
 *
 * This file is part of libcppsim
 *
 * Copyright (C) 2003, 2004 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/
#include <list>
#include <string>
#include <iostream>
#include "cppsim.hh"

extern "C" {
#include <getopt.h>
}

using namespace std;
using namespace cppsim;

//
// Forward Declarations
//
class job;
class server;
class sink;
class source;

//
// Simulation parameters
//
int numRep = 5;         // Number of replications, >1
double confp = 0.9;     // Confidence probability
double simLen = 1e6;    // Sim. Length in time units
bool doDebug = false;
int trLen = 0;          // Length (num of observations) in the transient 
sqs::sched_t schedType = sqs::SQSDLL;

//
// Entities
//
handle<server>  srv;
handle<sink>    snk;
handle<source>  src;

// 
// Counters, histograms and randoms
//
statistic*      wtime;          // job's waiting time
statistic*      tput;           // Throughput
rng*            job_inter_time; // Mean job interarrival time
rng*            job_serv_time;  // Mean job service time

//
// class SERVER
//
class server: public process {
public:
    server( const string& name ) : 
	process( name ),
        _c( new counter("Busy Time") ) 
    {
    
    };
    virtual ~server( ) 
    {
        delete _c;
    };
    list< handle<job> > queue;
protected:
    void inner_body( void );
    counter* _c;
};


//
// class SINK
//
class sink: public process {
public:
    sink( const string& name ) : 
	process( name ) { };
    list< handle<job> > queue;
protected:
    void inner_body( void );
};


//
// class SOURCE
//
class source: public process {
public:
    source( const string& name ) : 
	process( name ) { };
protected:
    void inner_body( void );
};


//
// class JOB
//
class job: public process {
public:
    job( const string& name ) : 
	process( name ),
	createdAt( time( ) ),
	servicedAt( 0.0 ) { };
    double createdAt;
    double servicedAt;
protected:
    void inner_body( void );
};


class simulation_main: public repSimulation {
public:
    simulation_main( double simDuration,
                     int maxReplications,
                     double relWidth );
    virtual ~simulation_main( ) { };
    void initialize( );
    void finalize( ) { };
};


//
// Body Definitions
//
void source::inner_body( void )
{    
    while ( 1 ) {
        hold( job_inter_time->value( ) );
        handle<job> j( new job( "Job" ) );
        j->activateAfter( current( ) );
    }
}

void job::inner_body( void )
{
    if ( srv->idle( ) ) 
        srv->activateAfter( current( ) );
    srv->queue.push_back( this );

    passivate( ); // wait for service

    wtime->update( servicedAt - createdAt );

    if ( snk->idle( ) )
	snk->activateAfter( current( ) );
    snk->queue.push_back( this );
}

void sink::inner_body( void )
{
    while ( 1 ) {
        if ( queue.empty( ) )
            passivate( );
        queue.pop_front( );
    }    
}

void server::inner_body( void )
{
    double t_stop    = 0.0;
    double t_restart = 0.0;
    int completed = 0;

    while ( 1 ) {
        if ( queue.empty( ) ) {
	    t_stop = time( );
            passivate( );
	    t_restart = time( );
	}
        handle<job> j( queue.front( ) );
        queue.pop_front( );
        j->servicedAt = time( );
        hold( job_serv_time->value( ) );
	j->activateAfter( current( ) );
        completed++;
        tput->update( completed / time( ) );
    }    
}

simulation_main::simulation_main( double simDuration,
                                  int maxReplications,
                                  double relWidth ) :
    repSimulation( simDuration, maxReplications, relWidth )
{

}

void simulation_main::initialize( void )
{
    srv = new server( "Server" );
    snk = new sink( "Sink" );
    src = new source( "Source" );
    
    src->activate( );
    srv->activate( );
    snk->activate( );
}

int main( int argc, char* argv[] )
{
    INIT_SEQUENCING;
    char opt;

    /*
     * Read the command line parameters
     */
    while ( -1 != ( opt = getopt( argc, argv, "c:hr:l:t:DS:" ) ) ) {
	switch (opt) {
	case 'c':
	    confp = atof( optarg );
	    break;
	case 'h':
	    cout << "Usage: " << argv[0] 
		 << " [-c <confp>] [-h] [-r <numrep>] [-l <siml>] [-t <trlen>] [-D] [-S]" << endl
		 << endl
		 << " -c <confp>  Sets the confidence probability to <confp> (def: 0.9)" << endl
		 << " -r <numrep> Sets the number of replications to <numrep> (def: 1)" << endl
		 << " -l <siml>   Sets the length of the simulation, in time units (def 1e6)" << endl
		 << " -t <trlen>  Length of the transient period (def 0)" << endl
		 << " -D          Turns on lots of debugging messages (def off)" << endl
		 << " -S          Sqs type (0 = Doubly linked List, 1 = Priority Queue)" << endl
		 << " -h          Displays this help" << endl
		 << endl;
	    exit( 0 );
	    break;
	case 'r':
	    numRep = atoi( optarg );
	    assert( numRep >= 1 );
	    break;
	case 'l':
	    simLen = atof( optarg );
	    break;
	case 't':
	    trLen = atoi( optarg );
	    break;
	case 'D':
	    doDebug = true;
	    process::getLogger()->enable( logger::LOG_INFO | logger::LOG_DEBUG );
	    break;
	case 'S':
	    schedType = (sqs::sched_t)atoi( optarg );
	    break;
	default:
	    cerr << "Unrecognized option: " << argv[ optind ] << endl;
	    exit( -1 );
	}
	
    }

    simulation* s = new simulation_main( simLen, numRep, 0.1 );
    s->setReport( new simpleReport( "M/M/1 Simulation Results" ) );

    //
    // Initialize the variables
    //
    s->registerStatistic( wtime = new repmean( "Waiting_Time", .9 ) );
    s->registerStatistic( tput = new repmean( "Throughput", .9 ) );
    s->registerRng( job_inter_time = new rngExp( "Job Interarr. Time", 1.0/0.09 ) );
    s->registerRng( job_serv_time = new rngExp( "Service Time", 1.0/0.1 ) );
    s->registerReportable( wtime );
    s->registerReportable( tput );

    s->run();
    s->getReport()->print();

    delete s;
    return 0;
}
