/****************************************************************************
 *
 * qn2.cc
 *
 * This file is part of libcppsim
 *
 * Copyright (C) 2005 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#include <cassert>
#include "qn2.hh"

vector< arrival* > _arrivals;
vector< server* > _servers;
fel _simulation;

//////////////////////////////////////////////////////////////////////////////
fel::fel( ) :
    _fel( )
{

}

void fel::run( void )
{
    while( !_fel.empty() ) {
        absEvent* e( _fel.top() );
        _fel.pop();
        assert( e->time( ) >= _t );
        _t = e->time( );
        e->execute( );
        delete e;
    }
}

void fel::scheduleAt( absEvent* e, double t )
{
    assert( t >= time() );
    e->setTime( t );
    _fel.push( e );
}

void fel::schedule( absEvent* e, double dt )
{
    e->setTime( time()+dt );
    _fel.push( e );
}

//////////////////////////////////////////////////////////////////////////////
server::server( int capacity, 
                vector< double > servrate, 
                vector< double > routing ) :
    _capacity( capacity ),
    _q( ),
    _blocked( ),
    _servTime( ),
    _nextServer( "next", routing ),
    _status( IDLE )
{
    vector< double >::iterator it;
    for ( it = servrate.begin(); it != servrate.end(); it++ ) {
        _servTime.push_back( new rngExp( "pippo", 1.0/ *it ) );
    }
    assert( _servTime.size() == servrate.size() );
};

server::~server( )
{

};

//////////////////////////////////////////////////////////////////////////////
//
// Nuovo arrivo esterno al server i di un nuovo job di classe c
class evExtArrival : public absEvent {
public:
    evExtArrival( int c, int s ) :
        absEvent( ), 
        _c( c ), 
        _s( s )
    { 

    };
    virtual ~evExtArrival( ) 
    {

    };
    void execute( void );
protected:
    int _c;
    int _s;
};

//////////////////////////////////////////////////////////////////////////////
//
// Completamento del servizio del primo job in coda al server i
class evComplete : public absEvent {
public:
    evComplete( int i ) : 
        absEvent( ),
        _i( i )
    { 

    };
    virtual ~evComplete( )
    {

    };
    void execute( void );
protected:
    int _i;
};

//////////////////////////////////////////////////////////////////////////////
//
// Arrivo interno su i
class evArrival : public absEvent {
public:
    evArrival( int i, job* j ) : 
        absEvent( ),
        _i( i ),
        _j( j )
    { 

    };
    virtual ~evArrival( )
    {

    };
    void execute( void );
protected:
    int _i;
    job* _j;
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void evExtArrival::execute( void )
{
    server* srv( _servers[ _s ] );
    if ( !srv->full() ) {
        job* j( new job( _c ) );
        absEvent* ev( new evArrival( _s, j ) );            
        _simulation.schedule( ev );
    }
    absEvent* ev2( new evExtArrival( _c, _s ) );
    _simulation.schedule( ev2, _arrivals[_s]->getTime() );
};


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void evArrival::execute( void )
{
    server* srv( _servers[_i] );
    srv->addJob( _j );
    if ( server::IDLE == srv->getStatus() ) {
        srv->setStatus( server::RUNNING );
        absEvent* ev( new evComplete( _i ) );            
        _simulation.schedule( ev, srv->servTime( _j->_c ) );
    }
};

//////////////////////////////////////////////////////////////////////////////

void evComplete::execute( void )
{
    server* srv( _servers[ _i ] );
    int j( srv->nextServer() );
    if ( j < 0 ) {
        // Caso 1. Il job esce dal sistema
        job* j( srv->removeJob( ) );
        delete j;
    } else {
        server* next( _servers[ j ] );
        if ( next->full() ) {
            // Case 2a. La destinazione e' piena
            srv->setStatus( server::BLOCKED );
            next->addBlocked( _i );
        } else {
            // Case 2b. La destinazione non e' piena

            // Sposta il job nella coda di destinazione
            job* jb( srv->removeJob( ) );   
            absEvent* ev( new evArrival( j, jb ) );
            _simulation.schedule( ev );
            
            // Processa il prossimo job
            if ( srv->empty() )
                srv->setStatus( server::IDLE );
            else {
                if ( !srv->emptyBlocked() ) {
                    int a( srv->removeBlocked() );
                    absEvent* ev( new evComplete( a ) );
                    _simulation.schedule( ev );
                }
                srv->setStatus( server::RUNNING );
                job* jb( srv->firstJob( ) );
                absEvent* ev( new evComplete( _i ) );
                _simulation.schedule( ev, srv->servTime( jb->_c ) );
            }
        }
    }
}

int main( void )
{
    vector< double > rate( 2 );
    rate[0] = 1.0;
    rate[1] = 2.0;
    vector< double > routing( 4 );
    routing[0] = 0.25;
    routing[1] = 0.25;
    routing[2] = 0.25;
    routing[3] = 0.25;

    _servers.resize( 4 );
    _servers[0] = new server( 4, rate, routing );
    _servers[1] = new server( 4, rate, routing );
    _servers[2] = new server( 4, rate, routing );
    _servers[3] = new server( 4, rate, routing );

    return 0;
}
