/****************************************************************************
 *
 * qn2.hh
 *
 * This file is part of libcppsim
 *
 * Copyright (C) 2005 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#ifndef QN2_HH
#define QN2_HH

#include <queue>
#include <functional>
#include <vector>
#include "cppsim.hh"

using namespace std;
using namespace cppsim;

//////////////////////////////////////////////////////////////////////////////

class absEvent {
public:
    virtual ~absEvent( ) { };
    virtual void execute( void ) = 0;
    double time( void ) const { return _time; };
    void setTime( double t ) { _time = t; };
protected:
    absEvent( double t = 0.0 ) : _time( t ) { };
    double _time;
};

//////////////////////////////////////////////////////////////////////////////

class fel {
public:
    fel( );
    virtual ~fel( ) { };
    void scheduleAt( absEvent* e, double t );
    void schedule( absEvent* e, double dt=0 );
    void run( void );
    double time( void ) const { return _t; };
protected:

    struct my_compare : public binary_function< absEvent*, absEvent*, bool >
    {
        bool operator()( absEvent*& e1, absEvent*& e2 )
        {
            return ( e1->time() < e1->time() );
        };
    };

    priority_queue< absEvent*, vector< absEvent* >, fel::my_compare > _fel;
    double _t;
};

//////////////////////////////////////////////////////////////////////////////

class arrival {
public:
    arrival( int c, int s, double r ) :
        _class( c ),
        _server( s ),
        _r( new rngExp( "arrival", 1.0/r ) )
    {
        
    };
    virtual ~arrival( )
    { 
        delete _r; 
    };
    // Accessors
    int getClass( void ) const { return _class; };
    int getServer( void ) const { return _server; };
    double getTime( void ) { return _r->value(); };
protected:
    int _class;
    int _server;
    rng* _r;
};

//////////////////////////////////////////////////////////////////////////////
typedef struct job {
public:
    job( int c ) : _c( c ) { };
    int _c;
} job;

//////////////////////////////////////////////////////////////////////////////
class server
{
public:
    typedef enum { IDLE, RUNNING, BLOCKED } status_t;

    server( int capacity, vector<double> servrate, vector<double> routing );
    virtual ~server( );
    // Accessors
    bool full( void ) { return (_q.size() >= _capacity); };
    bool empty( void ) { return _q.empty(); };
    job* firstJob( void ) { return _q.front(); };
    bool emptyBlocked( void ) { return _blocked.empty(); };
    status_t getStatus( void ) const { return _status; };
    void setStatus( status_t s ) { _status = s; };
    double servTime( int c ) { return _servTime[c]->value(); };
    int nextServer( void ) { return (int)_nextServer.value(); };
    // Modifiers
    void addJob( job* j ) { _q.push( j ); };
    job* removeJob( void ) { job* j( _q.front() ); _q.pop(); return j;};
    void addBlocked( int i ) { _blocked.push( i ); };
    int removeBlocked( void ) { int i( _blocked.front() ); _blocked.pop(); return i; };
protected:
    int _capacity;
    queue< job* > _q; // lista dei job in coda
    queue< int > _blocked; // lista dei serventi bloccati
    vector< cppsim::rng* > _servTime;
    cppsim::rngDiscEmpirical _nextServer;
    status_t _status;
};

#endif
