/****************************************************************************
 *
 * rng_test.cc
 *
 * Test suite for random number generators.
 * This file is part of libcppsim
 *
 * Copyright (C) 2003, 2004 Moreno Marzolla
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#include <vector>
#include <string>
#include <cmath>
#include <algorithm>
#include <functional>
#include "cppsim.hh"

using namespace std;
using namespace cppsim;

class rng_test {
public:
    rng_test( rng* t, // RNG to test
              unsigned int l = 10000   // How many numbers to draw
              ) :
        _t( t ),
        _c( t->makeCdf() ),
        _l( l )
    { };
    virtual ~rng_test( ) { delete _t; };
    void do_test( void ) throw( runtime_error );
protected:
    rng* _t;
    cdf* _c;
    unsigned int _l;
};

void rng_test::do_test( void ) throw( runtime_error )
{
    unsigned int i;
    mean*       _m( new mean( "mean::"+_t->name(), 0.95 ) );
    confInt     compMean;
    double      compVar;

    cout << "Testing " << _t->name() << "...";
    for ( i=0; i<_l; i++ ) {
        _m->update( _t->value( ) );
    }

    try {
        compMean = _m->value();
        compVar = _m->variance();
    } catch( runtime_error& e ) {
        cerr << "Runtime error caught: " 
             << e.what() << endl;
    }
    Assert( compMean.contains( _c->m() ),
            runtime_error( oform( "rng(\"%s\") produced a mean in (%f,%f) "
                                  "which does not contain the true value %f\n",
                                  _t->name().c_str(),
                                  compMean.lBound(),
                                  compMean.uBound(),
                                  _c->m() )
                           )
            );
    Assert( fabs( 1.0 - compVar / _c->v() ) < 1e-1 , // 10% relative error
            runtime_error( oform( "rng(\"%s\") produced a variance of %f "
                                  "which is not close to the true value %f\n",
                                  _t->name().c_str(),
                                  compVar,
                                  _c->v() )
                           )
            );
    cout << "Passed" << endl;
    delete _m;
};


int main( int argc, char* argv[] )
{
    vector< rng_test* > tests;

    // rngExp
    tests.push_back( new rng_test( new rngExp( "Exp(100)", 190.0 ) ) );
    tests.push_back( new rng_test( new rngExp( "Exp(0.1)", 0.1 ) ) );

    // rngNormal
    tests.push_back( new rng_test( new rngNormal( "Normal(10,20)", 10.0, 20.0 ) ) );
    tests.push_back( new rng_test( new rngNormal( "Normal(0,1)", 0.0, 1.0 ) ) );
    // rngDiscUniform
    tests.push_back( new rng_test( new rngDiscUniform( "Discrete_Uniform(0,10)", 0, 10 ) ) );

    // rngUniform
    tests.push_back( new rng_test( new rngUniform( "Uniform(0,10)", 0, 10 ) ) );
    tests.push_back( new rng_test( new rngUniform( "Uniform(-5,20)", -5, 10 ) ) );

    // rngBernoulli
    tests.push_back( new rng_test( new rngBernoulli( "Bernoulli(0.1)", 0.1 ) ) );
    tests.push_back( new rng_test( new rngBernoulli( "Bernoulli(0.99)", 0.99 ) ));

    // rngChisq
    tests.push_back( new rng_test( new rngChisq( "Chisq(100)", 100 ) ) );
    tests.push_back( new rng_test( new rngChisq( "Chisq(20)", 20 ) ) );
    tests.push_back( new rng_test( new rngChisq( "Chisq(5)", 5 ) ) );

    // rngBionomial
    tests.push_back( new rng_test( new rngBinomial( "Binomial(10,0.1)", 10, 0.1 ) ) ); 
    tests.push_back( new rng_test( new rngBinomial( "Binomial(100,0.9)", 100, 0.9 ) ) ); 
    
    // rngGamma
    tests.push_back( new rng_test( new rngGamma( "Gamma(0.5, 0.5)", 0.5, 0.5 ) ) );
    tests.push_back( new rng_test( new rngGamma( "Gamma(0.9, 0.9)", 0.9, 0.9 ) ) );
    tests.push_back( new rng_test( new rngGamma( "Gamma(10.0, 2.0)", 10.0, 2.0 ) ) );                                     

    //
    // Actually do the tests
    //
    try {
        for_each( tests.begin(), tests.end(),
                  mem_fun_t< void, rng_test >( & rng_test::do_test ) );
    } catch( runtime_error& e ) {
        cerr << "Runtime error caught: "
             << e.what() << endl;
        return -1;
    }

    return 0;
}
